//	compares two files and writes out lines in common and lines unique to each

/*
	File intersection program, by Jason Owen.  Written October 2001.
	Thanks to Paul Trout for sending me the source to his C version.
	
	This is kinda messy, I know, but I wrote it a long time ago.  I'm
	fairly sure there's a more efficient way to do this, but again, I
	wrote this a long time ago, and it works, so it's Good Enough.
	  
	See license.txt for usage and distribution requirements. You should
	probably give Paul Trout credit, as well, for the original idea. Plus
	(this part is optional), I'd be really interested in what you use it for,
	so please feel free to drop me a line at shannara256@yahoo.com saying how
	you're using it or how it's been useful. 
*/

#define VERSION "0.1.1"

#define CASESENSITIVE_DEFAULT false
#define LINENUMBERING_DEFAULT false
#define UNIQUE1_FILE_DEFAULT "unique1"
#define UNIQUE2_FILE_DEFAULT "unique2"
#define COMMON_FILE_DEFAULT  "common"

#include <iostream>
#include <fstream>

#include <getopt.h>

#include <string>
using namespace std;

/*	prints info relevent to errorlevel and exits
	str is either the program name (argv[0]), or related to the error
*/
void usage(char* str, int errorlevel)
{
	switch (errorlevel)
	{
	case 1:	//help
		cout << "Usage: " << str << " [OPTIONS]... file1 file2" << endl;
		cout << "Compares two files and outputs three files:" << endl
			 << "      UNIQUE1 - by default " << UNIQUE1_FILE_DEFAULT << ", contains lines found only in file1" << endl
			 << "      UNIQUE2 - by default " << UNIQUE2_FILE_DEFAULT << ", contains lines found only in file2" << endl
			 << "      COMMON - by default " << COMMON_FILE_DEFAULT << ", contains lines found in both file1 and file2" << endl;
		cout << endl;
		cout << "   -i  --ignore-case   ignore case when comparing lines" << (CASESENSITIVE_DEFAULT==0 ? ".  this is the default":"") << endl;
		cout << "   -m  --match-case    pay attention to case" << (CASESENSITIVE_DEFAULT==1 ? ".  this is the default":"") << endl;
		cout << "   -n  --numbering     print line numbers in output files" << (LINENUMBERING_DEFAULT==1 ? ".  this is the default":"") << endl;
		cout << "   -N  --no-numbering  don't print line numbers" << (LINENUMBERING_DEFAULT==0 ? ".  this is the default":"") << endl;
		cout << "   -1  --unique1 U1    use file U1 instead of " << UNIQUE1_FILE_DEFAULT << endl;
		cout << "   -2  --unique2 U2    use file U2 instead of " << UNIQUE2_FILE_DEFAULT << endl;
		cout << "   -c  --common C      use file C instead of " << COMMON_FILE_DEFAULT << endl;
		cout << "   -h  --help          show this screen" << endl;
		cout << "   -v  --version       show version information" << endl;
		cout << endl;
	case 0:	//version
		cout << "Original C program by Paul Trout (ptrout@usa.net)" << endl;
		cout << "Recreated in C++ and upgraded by Jason Owen (shannara256@yahoo.com)" << endl;
		cout << "version " << VERSION << " (using Linux kernel numbering scheme)" << endl;
		exit(0);

	//file errors
	case 10:	//not enough files listed
		cout << "You must specify two files to compare." << endl;
		cout << "See " << str << " --help for more information." << endl;
		exit(1);
	case 11:	//file open read failure
		cout << "Error opening " << str << " for reading." << endl;
		cout << "Check that it exists and is not being used." << endl;
		exit(2);
	case 12:	//file open write failure
		cout << "Error opening " << str << " for writing." << endl;
		cout << "Check that it is not being used and is a valid filename." << endl;
		exit(3);
	case 13:	//processing infile error
		cout << "Error processing " << str << '.' << endl;
		cout << "Are you sure it's a text file?" << endl;
		exit(4);
	case 20:
		cout << "Out of memory." << endl;
		cout << "Maybe try smaller files?" << endl;
		exit(5);
	}
}


/*	options:	-i --ignore-case / -m --match-case, -n --numbering / -N --no-numbering
				-1 --unique1: output file for lines unique to file1
				-2 --unique2: output file for lines unique to file2
				-c --common:  output file for lines common to both files
  */

static char short_options[] =
"imnN1:2:c:hv";

static struct option long_options[] =
{
	{"ignore-case",		no_argument,		NULL, 'i'},
	{"match-case",		no_argument,		NULL, 'm'},
	{"numbering",		no_argument,		NULL, 'n'},
	{"no-numbering",	no_argument,		NULL, 'N'},
	{"unique1",			required_argument,	NULL, '1'},
	{"unique2",			required_argument,	NULL, '2'},
	{"common",			required_argument,	NULL, 'c'},

	{"help",			no_argument,		NULL, 'h'},
	{"version",			no_argument,		NULL, 'v'}
};


//process options from command line
void getoptions(int argc, char* argv[],
				bool &casesensitive, bool &linenumbering,
				char* &file1, char* &file2,
				char* &unique1, char* &unique2, char* &common)
{
	int current;
	casesensitive = CASESENSITIVE_DEFAULT;
	linenumbering = LINENUMBERING_DEFAULT;
	unique1 = UNIQUE1_FILE_DEFAULT;
	unique2 = UNIQUE2_FILE_DEFAULT;
	common = COMMON_FILE_DEFAULT;

	while ((current=getopt_long(argc, argv, short_options, long_options, NULL))!=-1)
	{
		switch (current)
		{
		case 'i':	casesensitive=false;	break;
		case 'm':	casesensitive=true;		break;

		case 'n':	linenumbering=true;		break;
		case 'N':	linenumbering=false;	break;

		case '1':	unique1=optarg;			break;
		case '2':	unique2=optarg;			break;
		case 'c':	common =optarg;			break;


		case 'h':	usage(argv[0], 1);		break;
		case 'v':	usage(argv[0], 0);		break;
		}
	}

	if (optind < argc-1)
	{
		file1 = argv[optind];
		file2 = argv[++optind];
	}
	else
		usage(argv[0], 10);
}

struct fileline		//singly linked list of lines
{
	unsigned long	linenum;
	unsigned long	linenum2;
	string			line;
	fileline*		next;
};

bool is_line_notblank(string line)
{
  unsigned int c;
  bool fnd=false;		                         // Non space or control found

  for (c=0; c<line.length() && !fnd; c++)
  {
	  if (!((isspace(line[c]) || iscntrl(line[c]))))	//if whitespace or ctrl char
		  fnd=true;										//then we found something
  }
  
  return fnd;
}

//frees a linked list starting at head
void freelist(fileline* head)
{
	fileline* prev;
	while (head->next != NULL)
	{
		prev=head;
		head=head->next;
		delete prev;
	}
	delete head;
}

//creates a linked list containing all the lines in the file, and returns the first
fileline* make_list(fstream &filename)
{
	string buffer;
	fileline* ret = new fileline;
	if (ret == NULL)	usage("", 20);
	fileline* current = ret;
	int c=1;	//line counter
	
	while (!filename.eof())
	{
		getline(filename, buffer);
		if (is_line_notblank(buffer))
		{
			current->linenum = c;
			current->linenum2 = 0;
			current->line = buffer;
			current->next = new fileline;
			if (current->next == NULL)
			{
				freelist(ret);
				usage("", 20);
			}
			current=current->next;
		}
		c++;
	}
	current->next = NULL;	//terminate the list
	return ret;
}

//goes through list to find search
fileline* find_line(fileline* search, fileline* list, bool casesensitive)
{
	if (casesensitive)
	{
		while (list->next != NULL)
		{
			if (!strcmp(list->line.c_str(), search->line.c_str()))
				return list;
			list=list->next;
		}
	}
	else
	{
		while (list->next != NULL)
		{
			if (!stricmp(list->line.c_str(), search->line.c_str()))
				return list;
			list=list->next;
		}
	}
	return NULL;
}

//if you want to speed this program up, this would probably be the best thing to replace
void mark_common(fileline* list1, fileline* list2, bool casesensitive)
{
	fileline* current = list2, *finder;
	while (current->next != NULL)
	{
		if ((finder = find_line(current, list1, casesensitive)) != NULL)
		{
			finder->linenum2 = current->linenum;
		}
		current=current->next;
	}

	current=list1;
	while (current->next != NULL)
	{
		if ((finder = find_line(current, list2, casesensitive)) != NULL)
		{
			finder->linenum2 = current->linenum;
		}
		current=current->next;
	}
}

int main(int argc, char* argv[])
{
	bool casesensitive, linenumbering;
	char* file1name, *file2name, *unique1name, *unique2name, *commonname;
	getoptions(argc, argv, casesensitive, linenumbering, \
			   file1name, file2name, unique1name, unique2name, commonname);

	//open files for reading
	fstream file1(file1name, ios::in);
	if (!file1.is_open())	usage(file1name, 11);
	fstream file2(file2name, ios::in);
	if (!file2.is_open())	usage(file2name, 11);

	//open files for writing
	fstream unique1(unique1name, ios::out);
	if (!unique1.is_open())	usage(unique1name, 12);
	fstream unique2(unique2name, ios::out);
	if (!unique2.is_open())	usage(unique2name, 12);
	fstream common(commonname, ios::out);
	if (!common.is_open())	usage(commonname, 12);

	//create lists
	fileline* list1 = NULL;
	if ((list1 = make_list(file1)) == NULL)	usage(file1name, 13);
	fileline* list2 = NULL;
	if ((list2 = make_list(file2)) == NULL) usage(file2name, 13);

	//compare lists
	mark_common(list1, list2, casesensitive);

	//write output to files
	fileline* current=list1;
	unique1 << "Lines unique to " << file1name << endl;
	unique2 << "Lines unique to " << file2name << endl;
	common << "Lines common to both " << file1name << " and " << file2name << endl;
	if (linenumbering)	//this is way up here instead of nested farther down in the loop to speed things up a bit
	{
		common << "First number is the line from " << file1name << endl;
		common << "The second is the line number from " << file2name << endl;
		while (current->next != NULL)
		{
			if (current->linenum2)	common << current->linenum << ", " << current->linenum2 << '\t' << current->line << endl;
			else	unique1 << current->linenum << '\t' << current->line << endl;

			current=current->next;
		}
		current=list2;
		while (current->next != NULL)
		{
			if (!current->linenum2)	unique2 << current->linenum << '\t' << current->line << endl;
			
			current=current->next;
		}
	}
	else	//no line-numbering
	{
		while (current->next != NULL)
		{
			if (current->linenum2)	common << current->line << endl;
			else	unique1 << current->line << endl;

			current=current->next;
		}
		current=list2;
		while (current->next != NULL)
		{
			if (!current->linenum2)	unique2 << current->line << endl;
			current=current->next;
		}
	}

	return 0;
}
